import pyopencl as cl
import numpy,struct,math,os,sys

block_size = 10
devs = cl.get_platforms()[0].get_devices()
try:
	ctx = cl.Context(devices = devs[:-1])
except:
	ctx = cl.Context(devices = devs[-1:])
queue = cl.CommandQueue(ctx)
prg = cl.Program(ctx, """
#define BLOCK_SIZE %(block_size)i
#pragma OPENCL EXTENSION cl_khr_fp64: enable

__kernel __attribute__((reqd_work_group_size(BLOCK_SIZE,BLOCK_SIZE,1)))
	void nw(__global const float *seq, __global const uint *ind, __global float *out, unsigned int numSeq,float nu,uint dim)
    {
	int gx = get_global_id(0);
	int gy = get_global_id(1);
      	if(gx<numSeq && gy<numSeq){
		float create,modify;
		float crea,modi,dele;
		float tmpBuff[2000];
		for (uint i=0;i<ind[gy];i++){
			tmpBuff[i]= 0;
		}
		uint indI, indJ, i, j,indIK,indJK;
		float sum;
		for (i=1;i<ind[gx];i++){
			indI = ind[gx+numSeq]+i*dim;
	   		modify = 0;
			create = 0;
			for (j=1;j<ind[gy];j++){
				indJ = ind[gy+numSeq]+j*dim;
				sum = 0.0f;
				for (int k=0;k<dim;k++){
					indIK = indI+k;
					indJK = indJ+k;
					sum += (fabs(seq[indIK]-seq[indJK])+fabs(seq[indIK-dim]-seq[indJK-dim]))+3.0f*fabs((seq[indIK]-seq[indIK-dim])-(seq[indJK]-seq[indJK-dim]));
				}
				create = max(max(modify+exp(-nu*sum),tmpBuff[j]),create);
				modify = tmpBuff[j];
		  		tmpBuff[j] = create;  
			}
		}
		out[gx*numSeq+gy] = tmpBuff[ind[gy]-1]/sqrt((float)ind[gx]*ind[gy]);
	}
}

__kernel __attribute__((reqd_work_group_size(BLOCK_SIZE,BLOCK_SIZE,1)))
	void sw(__global const float *seq, __global const uint *ind, __global float *out, unsigned int numSeq,float gap,float nu,uint dim)
    {
	int gx = get_global_id(0);
	int gy = get_global_id(1);
      	if(gx<numSeq && gy<numSeq){
		float create,modify;
		float crea,modi,dele;
		float tmpBuff[2000];
		for (uint i=0;i<ind[gy];i++){
			tmpBuff[i]= 0;
		}
		uint indI, indJ, i, j,indIK,indJK;
		float maxV = 0.0f;
		float sum;
		for (i=1;i<ind[gx];i++){
			indI = ind[gx+numSeq]+i*dim;
	   		modify = 0;
			create = 0;
			for (j=1;j<ind[gy];j++){
				indJ = ind[gy+numSeq]+j*dim;
				sum = 0.0f;
				for (int k=0;k<dim;k++){
					indIK = indI+k;
					indJK = indJ+k;
					sum += (fabs(seq[indIK]-seq[indJK])+fabs(seq[indIK-dim]-seq[indJK-dim]))+3.0f*fabs((seq[indIK]-seq[indIK-dim])-(seq[indJK]-seq[indJK-dim]));
				}
				create = max(max(max(modify+1.0f-nu*sum,tmpBuff[j]+gap),create+gap),0.0f);
				maxV = max(maxV,create);
				modify = tmpBuff[j];
		  		tmpBuff[j] = create;  
			}
		}
		out[gx*numSeq+gy] = maxV;
	}
}
"""%{"block_size":block_size}).build()

nw = prg.nw
nw.set_scalar_arg_dtypes([None, None, None, numpy.uint32, numpy.float32, numpy.uint32])
sw = prg.sw
sw.set_scalar_arg_dtypes([None, None, None, numpy.uint32, numpy.float32,numpy.float32, numpy.uint32])


def readDB(fileName,local=True):
	t = time()
	data = open(fileName)
	dim = struct.unpack('I',data.read(4))[0]
	totLen = struct.unpack('I',data.read(4))[0]
	numSeq = struct.unpack('I',data.read(4))[0]
	indices = numpy.empty(numSeq*2, dtype=numpy.uint32)
	values = numpy.empty(totLen, dtype=numpy.float32)
	for i in range(0,numSeq):
		indices[i] = struct.unpack('I',data.read(4))[0]
	accu = 0
	for i in range(0,numSeq):
		indices[i+numSeq] = accu
		accu += indices[i]*dim
	for i in range(0,totLen):
		values[i] = struct.unpack('f',data.read(4))[0]
	if local:
		for i in range(0,numSeq):
			for k in range(0,dim):
				avg = 0.0
				for j in range(0,indices[i]):
					avg += values[indices[i+numSeq]+k+j*dim]
				avg /= indices[i]
				for j in range(0,indices[i]):
					values[indices[i+numSeq]+k+j*dim] = values[indices[i+numSeq]+k+j*dim]/avg
	return (dim,numSeq,indices, values)


from time import time

nuVals = [0.67,0.41,0.24,0.15];
gapVals = [-0.53,-0.5];

t = time()
if len(sys.argv)==4 and os.path.isfile(sys.argv[1]):
	if sys.argv[3] == '-sw':
		(dim,numSeq,indices,sequences) = readDB(sys.argv[1])
	elif sys.argv[3] == '-nw':
		(dim,numSeq,indices,sequences) = readDB(sys.argv[1],False)
	else:
		print "Usage is \"python lna.py DBname outputFile -method\" with method = (nw or sw)"
		exit(1)
else:
	print "Usage is \"python lna.py DBname outputFile -method\" with method = (nw or sw)"
	exit(1)

out = open(sys.argv[2],'w')
print 'db loading : ',time()-t
result = numpy.empty((numSeq, numSeq)).astype(numpy.float32)
mf = cl.mem_flags
s_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=sequences)
i_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=indices)
out_buf = cl.Buffer(ctx, mf.WRITE_ONLY , result.nbytes)
print 'loading + sending data to GPU : ',time()-t
if sys.argv[3] == '-sw':
	exec_evt = sw(queue,((numSeq/block_size+1)*block_size, (numSeq/block_size+1)*block_size),(block_size, block_size),s_buf, i_buf,out_buf,numSeq,gapVals[dim-1],nuVals[dim-1],dim)
elif sys.argv[3] == '-nw':
	exec_evt = nw(queue,((numSeq/block_size+1)*block_size, (numSeq/block_size+1)*block_size),(block_size, block_size),s_buf, i_buf,out_buf,numSeq,nuVals[1+dim],dim)
exec_evt.wait()
cl.enqueue_read_buffer(queue, out_buf, result).wait()
print 'Time after GPU computation : ',time()-t

for i in range(0,numSeq):
	for j in range(0,numSeq):
		if i!=j:
			out.write(str(result[i][j])+'\t')
	out.write('\n')

print 'Total time : ',time()-t
out.close()
