/*
 *  **************************************************************************
 *
 *  (c) 2012 Nicolas Bonnel
 * 
 *  **************************************************************************
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,  but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details. 
 *  
 *  You should have received a copy of the GNU Lesser General Public License 
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  **************************************************************************
 * 
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <cstring>

static union
{
	double d;
	struct
	{
#ifdef LITTLE_ENDIAN
		int j, i;
#else
		int i, j;
#endif
	} n;
} _eco;

#define EXP_A (1048576/M_LN2)	/* 2^20/LN(2) use 1512775 for integer version */
#define EXP_C 60801	 /* see text for choice of c values */
#define EXP(y) (_eco.n.i = EXP_A*(y) + (1072693248 - EXP_C), _eco.d)

using namespace std;
const float nuVals[] = {0.67f,0.41f,0.24f,0.15f};
const float gapVals[] = {-0.53f,-0.5f};

float nw(unsigned int dimension,float *seq,float *quer, unsigned int sOffset, unsigned int sLength, unsigned int qOffset, int qLength,float nu){
	float create,modify,res;
	float * tmpBuff = (float *)malloc(sizeof(float)*qLength);;
	for (int i=0;i<qLength;i++)
		tmpBuff[i]= 0;
	unsigned int indI, indJ, indIK,indJK;
	float sum;
	unsigned int i,j,k;
	for (i=1;i<sLength;i++){
		indI = (sOffset+dimension*i);
   		modify = 0;
		create = 0;
		for (j=1;j<qLength;j++){
			indJ = (qOffset+dimension*j);
			sum = 0.0f;
			for (k=0;k<dimension;k++){
				indIK = indI+k;
				indJK = indJ+k;
				sum += (abs(seq[indIK]-quer[indJK])+abs(seq[indIK-dimension]-quer[indJK-dimension]))+3.0f*abs((seq[indIK]-seq[indIK-dimension])-(quer[indJK]-quer[indJK-dimension]));
			}
			create = max(max(modify+(float)EXP(-nu*(sum)),tmpBuff[j]),create);
			modify = tmpBuff[j];
	  		tmpBuff[j] = create;  
		}
	}
	res = tmpBuff[qLength-1]/sqrt((float)(sLength-1)*(qLength-1));
	free(tmpBuff);
	return res;
}

float sw(unsigned int dimension,float *seq,float *quer, unsigned int sOffset, unsigned int sLength, unsigned int qOffset, int qLength,float nu,float gap){
	float create,modify;
	float * tmpBuff = (float *)malloc(sizeof(float)*qLength);;
	for (int i=0;i<qLength;i++)
		tmpBuff[i]= 0.0f;
	unsigned int indI, indJ,indIK,indJK;
	float sum;
	float result = 0.0f;
	unsigned int i,j,k;
	for (i=1;i<sLength;i++){
		indI = (sOffset+dimension*i);
   		modify = 0.0f;
		create = 0.0f;
		for (j=1;j<qLength;j++){
			indJ = (qOffset+dimension*j);
			sum = 0.0f;
			for (k=0;k<dimension;k++){
				indIK = indI+k;
				indJK = indJ+k;
				sum += (abs(seq[indIK]-quer[indJK])+abs(seq[indIK-dimension]-quer[indJK-dimension]))+3.0f*abs((seq[indIK]-seq[indIK-dimension])-(quer[indJK]-quer[indJK-dimension]));
			}
			create = max(max(max(modify+1.0f-nu*sum,tmpBuff[j]+gap),create+gap),0.0f);
			modify = tmpBuff[j];
	  		tmpBuff[j] = create;  
			result = max(result,create);
		}
	}
	free(tmpBuff);
	return result;
}

void divideByAvg(unsigned int dimension,float *seq,unsigned int sOffset, unsigned int sLength){
	for (unsigned int k=0;k<dimension;k++){
		float avg = 0.0f;
		for (unsigned int i=0;i<sLength;i++)
			avg += seq[sOffset+k+i*dimension];
		avg /= sLength;
		for (unsigned int i=0;i<sLength;i++)
			seq[sOffset+k+i*dimension] = seq[sOffset+k+i*dimension]/avg;
	}
}

void processDB(char *DBname,char *method){
	ifstream file(DBname, ios::binary);
	if (file){
		unsigned int dimension,totLen, numSeq;
		file.read((char*)&dimension, sizeof(unsigned int));
		file.read((char*)&totLen, sizeof(unsigned int));
		file.read((char*)&numSeq, sizeof(unsigned int));
		unsigned int *indices = (unsigned int *)malloc(sizeof(unsigned int)*numSeq*2);
		float *values= (float *)malloc(sizeof(float)*totLen);
		file.read((char*)indices, sizeof(unsigned int)*numSeq);
		unsigned int accu = 0;
		for (int i=0;i<numSeq;i++){
			indices[numSeq+i] = accu;
			accu += indices[i]*dimension;
		}
		file.read((char*)values, sizeof(float)*totLen);
		if (strcmp(method, "-sw")==0){
			for (int i=0;i<numSeq;i++)
				divideByAvg(dimension,values,indices[numSeq+i],indices[i]);
			for (int i=0;i<numSeq;i++){
				for (int j=i+1;j<numSeq;j++)
					cout << sw(dimension,values,values,indices[numSeq+j],indices[j],indices[numSeq+i],indices[i],nuVals[dimension-1],gapVals[dimension-1]) << "\t";
				cout << endl;
			}
		}else
			for (int i=0;i<numSeq;i++){
				for (int j=i+1;j<numSeq;j++)
					cout << nw(dimension,values,values,indices[numSeq+j],indices[j],indices[numSeq+i],indices[i],nuVals[1+dimension]) << "\t";
				cout << endl;
			}
		free(indices);
		free(values);
	}else
		cout << "Error, DB file not found"<<endl; 	
}



int main(int argc, char *argv[]){
	if (argc != 3 or (strcmp(argv[2], "-nw")!=0 and strcmp(argv[2], "-sw")!=0)){
		cout << "Usage is \"lnaAvA DBname -method\" with method = (nw or sw)"<<endl;
		exit(1);
	}
	processDB(argv[1],argv[2]);
	return 0;
}

